from .camera_.camera import Camera
from .camera_.camera_calibration import CameraCalibration
from .camera_.camera_settings import CameraSettings
from .connections import *
from .connections.network_location import NetworkLocation
from .connections.network_tables_connection import NetworkTablesConnection
from .connections.serial_connection import SerialConnection
from .contour_filters_.conditional_contour_filter import conditional_contour_filter
from .contour_filters_.contour_filter import contour_filter
from .contour_filters_.contour_filters import *
from .contour_filters_.sorters import *
from .directions_.directing_functions import *
from .directions_.direction_monitors.direction_monitor import DirectionMonitor
from .directions_.direction_monitors.stop_if_close_monitor import StopIfCloseMonitor
from .directions_.director import Director
from .display_image.display_contours import display_contours
from .display_image.display_image import display_image
from .exceptions_ import exceptions
from .hsv_calibration_.hsv_calibration import HSVCalibration
from .image_filters_.image_filters import *
from .image_utilities import drawing
from .image_utilities.naming_conventions import *
from .image_utilities.photo_array import photo_array
from .math_ import geometry
from .morphological_functions_.morphological_functions import erosion, dilation
from .thresholds.canny_edge import CannyEdge
from .thresholds.color_.built_in_colors import *
from .thresholds.color_.color import Color
from .thresholds.color_.multi_color import MultiColor
from .thresholds.threshold import Threshold
from .visions.ambient_vision import AmbientVision
from .visions.multi_vision import MultiVision
from .visions.vision import Vision

__version__ = '2020.1.4'



# Copyright 2018 - 2019 Ori Ben-Moshe - All rights reserved.
from functools import reduce
import time
import math
import copy
import types
import cv2
import numpy as np
import json
from typing import List, Union, Tuple, Any

from ..thresholds.color_ import built_in_colors
from ..exceptions_.exceptions import InvalidCustomFunctionError, CameraError
from ..camera_.camera import Camera
from ..partials.filter_applier import filter_applier
from ..thresholds.threshold import Threshold
from ..thresholds.color_.color import Color
from ..connections.connection import Connection
from ..directions_.director import Director
from ..camera_.camera_settings import CameraSettings
from ..connections.network_location import NetworkLocation
from ..directions_.directing_functions import center_directions


class Vision:
    """
    The Vision object represents the specifications to detect an object in an image.
    It performs the detection using a Threshold and filter functions.

    Additional capabilities and tuning options are:
        Image filters (Blurs, rotations, cropping),
        Morphological functions,
        Ovl color HSVCalibration,


    """
    def __init__(self, threshold: Threshold = None, contour_filters: List[types.FunctionType] = None,
                 director: Director = None, width=320, height=240, connection: Connection = None,
                 camera: Union[int, str, Camera, cv2.VideoCapture, Any] = None,
                 camera_settings: CameraSettings = None, morphological_functions: List[types.FunctionType] = None,
                 image_filters: List[types.FunctionType] = None, ovl_camera: bool = False, calibration: str = None):
        """
        The object that represents the pipeline of processing, detection, direction of values
        A connection object can be passed
        :param threshold: threshold object, creates the binary mask from a given image
        :param contour_filters: the list of contour_filter functions that
                                remove contours that aren't the target(s)
        :param director: a functions that receives a list or a single contour and returns director
        :param width: the width (in pixels) of images taken with the camera
        :param height: the height (in pixels)
        :param connection: a connection object that passes the result to the connection target
        :param camera: a Camera object (cv2.VideoCapture, ovl.Camera) or source from which to open a camera
        :param camera_settings: Special camera settings like calibration or offset used for
                                image correction and various direction calculations.
        :param morphological_functions: morphological functions used to apply on the binary
                               mask generated by the Threshold.
        :param image_filters: a list of image altering functions that are applied on the image.
        :param ovl_camera: a boolean that makes the camera opened to be ovl.Camera instead of cv2.VideoCapture
        :param calibration: a dictionary containing color calibration (HSVCalibration) coefficients and intercepts,
                            used for applying the calibration on the vision object
        """
        self.width = width
        self.height = height
        self.threshold = threshold
        self.contour_filters = contour_filters or []
        self.director = director or Director(center_directions, failed_detection=9999, target_amount=1)
        self.connection = connection
        self.image_filters = image_filters or []
        self.morphological_functions = morphological_functions or []
        self.camera = None
        self.camera_port = None
        self.camera_settings = camera_settings
        if isinstance(camera, (cv2.VideoCapture, Camera)):
            self.camera = camera
        elif camera is None:
            pass
        else:
            self.camera_setup(camera, width, height, ovl_camera=ovl_camera)

        self.calibration_path = None
        if calibration:
            self.calibration_path = calibration
            calibration = json.load(calibration)
            self.saturation_weight = calibration['saturation'] if 'saturation' in calibration else None
            self.brightness_weight = calibration['brightness'] if 'brightness' in calibration else None

    def __repr__(self):
        return str(self)

    def __str__(self):
        filters = [filter_function.__name__ for filter_function in self.contour_filters]
        threshold = self.threshold
        return "Vision: \n Threshold: {} \n Filters: {}".format(threshold, filters)

    @property
    def target_amount(self):
        """
        The wanted amount of targets
        (0 None or math.inf if there is no limit, 1 if 1 target is wanted etc.)
        """
        if self.director is None:
            return math.inf
        return self.director.target_amount

    def apply_morphological_functions(self, mask, morphological_functions=None):
        """
        Applies all morphological functions on the mask
        :param mask: the mask on which the functions should be applied
        :param morphological_functions: list of morphological_functions to be
                             applied instead of self.morphological_functions
        :return: the applied mask
        """
        if type(self.morphological_functions) not in (tuple, list, set):
            return mask
        morphological_functions = morphological_functions or self.morphological_functions
        return reduce(filter_applier, morphological_functions, mask)

    def send(self, data, *args, **kwargs) -> Union[None, Any]:
        """
        Sends data to the destination.
        :param data: The data to send to the Connection
        :param args: any other arguments for the send function in your connection
        :param kwargs: any other named arguments for the connection object
        :return: True if data was successfully sent, False if not.
        """
        return self.connection.send(*args, **kwargs, data=data)

    def send_to_location(self, data, network_location: NetworkLocation, *args, **kwargs):
        return self.connection.send_to_location(data, network_location, *args, **kwargs)

    def get_image(self) -> np.ndarray:
        """
        Gets an image from self.camera and applies image filters
        :return: the image, false if failed to get it
        """
        if self.camera is None:
            raise CameraError("No camera given, (Camera is None)")
        if not self.camera.isOpened():
            raise CameraError("Camera given is not open (Has it been closed or disconnected?)")
        output = self.camera.read()
        if len(output) == 2:
            ret, image = output
            return image if ret else False
        else:
            return output

    def get_filtered_image(self):
        """
        Gets an image from self.camera and applies all image filters
        :return: the image filter applied image
        """
        output = self.get_image()
        return self.apply_image_filters(output)

    def apply_filter(self, filter_function, contours, verbose=False):
        """
        Applies a filter function of the contour list
        :param filter_function: Filter functions are function with a contour list variable that apply some
        sort of filter on the contours, thus removing ones that don't fit the limit given by the filter.
        for example: straight_rectangle_filter removes contours that are not rectangles that are parallel to the frame
        of the picture
        :param contours: the contours on which the filter should be applied (list of numpy.ndarrays)
        :param verbose: if true_shape does not print anything
        :return: returns the output of the filter function.
        """
        if verbose:
            print('Before "{}": {}'.format(filter_function.__name__, len(contours)))
        filter_function_output = filter_function(contours)

        if isinstance(filter_function_output, tuple):
            if len(filter_function_output) == 2:
                filtered_contours, ratio = filter_function_output[0], filter_function_output[1]
            else:
                raise InvalidCustomFunctionError('Filter function must return between 1 and 2 lists.'
                                                 'Please refer to the Documentation: '
                                                 'https://github.com/1937Elysium/Ovl-Python')
        elif isinstance(filter_function_output, list):
            filtered_contours, ratio = filter_function_output, []
        else:
            raise TypeError('The contour list must be a list or tuple of 2 lists (contours and ratios)')
        return filtered_contours, ratio

    def apply_all_filters(self, contours: List[np.ndarray], verbose=False
                          ) -> Tuple[List[np.ndarray], List[float]]:
        """
         Applies all of the filters on a list of contours
        :param contours: List of contours (numpy arrays) to
        :param verbose: prints out information about filtering process if true (useful for debugging)
        :return: a list of all of the ratios given by the filter function in order.
        """
        ratios = []
        for filter_func in self.contour_filters:
            contours, ratio = self.apply_filter(filter_func, contours, verbose=verbose)
            ratios.append(ratio)
        if verbose:
            print("After all filters: {}".format(len(contours)))
        return contours, ratios

    def apply_image_filters(self, image: np.ndarray) -> np.ndarray:
        """
         applies all given image filters to the given image
        :param image: the image that the image filters should be applied on (numpy array)
        :return: the image with the filters applied
        """
        return reduce(filter_applier, self.image_filters, image)

    def apply_threshold(self, image: np.ndarray, threshold=None) -> np.ndarray:
        """
         gets a mask for a given img and Threshold object (uses self.Threshold if given threshold was none)
        :param image: the numpy array of the image
        :param threshold: the Threshold object used to create the binary mask
        :return: the binary mask
        """
        threshold = threshold or self.threshold
        return threshold.convert(image)

    def find_contours_in_mask(self, mask: np.ndarray, return_hierarchy=False, apply_morphs=True
                              ) -> List[np.ndarray]:
        """
         gets contours from the given mask and apply
        :param mask: binary image (mask), a numpy array
        :param return_hierarchy: if the hierarchy should be returned
        :param apply_morphs: if the morphological functions should be applied.
        :return: the list of contours
        """
        mask = self.apply_morphological_functions(mask) if apply_morphs else mask
        result = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        if len(result) == 3:
            _, contours, hierarchy = result
        elif len(result) == 2:
            contours, hierarchy = result
        else:
            raise ValueError("Invalid output from cv2.findContours, check that your cv2 (OpenCV) version is supported")
        return (contours, hierarchy) if return_hierarchy else contours

    def find_contours(self, image: np.ndarray, threshold=None, return_hierarchy=False) -> List[np.ndarray]:
        """
         Gets a list of all the contours within the threshold that was given
        :param threshold: the threshold that determines the binary masks
        :param image: image from which to get the contours
        :param return_hierarchy: if the hierarchy should be returned
        :return: list of all contours matching the range of hsv colours
        :rtype: list
        """
        threshold = threshold or self.threshold
        image_mask = self.apply_threshold(image, threshold=threshold)
        return self.find_contours_in_mask(image_mask, return_hierarchy=return_hierarchy)

    def get_directions(self, contours: List[np.ndarray], image: np.ndarray, sorter=None):
        """
        Calculates the directions, based on contours found in the given image
        :param contours: final contours after filtering
        :param image: the image from which to find the contours
        :param sorter: optional parameter, applies a sorter on the given contours
        :return: a string of the director (output of the director function),
                 length depends on the director function
        """
        return self.director.direct(contours, image, camera_settings=self.camera_settings, sorter=sorter)

    def camera_setup(self, source=0, image_width=None, image_height=None, ovl_camera=False):
        """
         Opens up the camera reference and sets a given width and height to all images taken
        :param image_width: the width of the images to be taken, 0 does not set a width
        :param image_height: the height of the images to be taken, 0 does not set a height
        :param source: the location from which to open the camera
                       string for network connections int for local USB connections.
        :param ovl_camera: if the camera object should be ovl.Camera
        :return: the camera object, also sets self.camera to the object.
        """
        image_height = image_height or self.height
        image_width = image_width or self.width
        self.camera_port = source
        if ovl_camera:
            robot_cam = Camera(source=source, image_width=image_width, image_height=image_height)
        else:
            robot_cam = cv2.VideoCapture(source)
            if image_width != -1:
                robot_cam.set(3, image_width)
            if image_height != -1:
                robot_cam.set(4, image_height)

        if not robot_cam.isOpened():
            raise CameraError("Camera did not open correctly! Camera source: {}".format(self.camera_port))
        self.camera = robot_cam
        return robot_cam

    def apply_on_sample(self, image: Union[np.ndarray, str], threshold: Threshold = None, display_result: bool = False,
                        delay=0, result_color: Color = built_in_colors.RED_HIGH_HSV, return_hierarchy=False
                        ) -> Tuple[List[np.ndarray], np.ndarray]:
        """
         Finds contours and applies filters on a single image given by a image path or image object (numpy array)
        :param image: an image on which to apply the vision object
        :param threshold: a custom threshold can be passed
        :param display_result: if the contours found should be displayed
        :param delay: waitKey delay (0 if not in a loop, non-zero if it is)
        :param result_color: the colour of the contour outline
        :param return_hierarchy: if the contour hierarchies should be return
        :return: the image (numpy array) and the contours found (list of numpy arrays)
                 Can also return the hierarchies if return_hierarchy is true
        """
        threshold = threshold or self.threshold
        image = cv2.imread(image) if type(image) == str else image
        contours, hierarchy = self.find_contours(image, threshold, return_hierarchy)
        if display_result:
            image_for_display = copy.copy(image)
            cv2.drawContours(image_for_display, contours, -1, result_color)
            cv2.imshow(str(time.time()), image_for_display)
            cv2.waitKey(delay)
        return (contours, image, hierarchy) if return_hierarchy else (contours, image)

    def detect(self, image, verbose=False, return_ratios=False):
        """
        Gets contours and applies all filters and returns the result, thus detecting the object according
        to the specifications in the vision
        :param image: image in which the vision should detect an object
        :param verbose: If information about the filtering should be printed
        :param return_ratios: if the ratios from the filters should be returned
        :return: contours and the filtered image and the ratios if return_ratios is true
        """
        image = self.apply_image_filters(image)
        contours = self.find_contours(image)
        contours, ratios = self.apply_all_filters(contours, verbose)
        return (contours, image, ratios) if return_ratios else (contours, image)



# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
import cv2
import numpy as np

from .color import Color
from ..threshold import Threshold


class MultiColor(Threshold):
    """
    Multi Color is like the Color object but it allows for the use of multiple ranges
    a main use is for representing the red color range.

    Example Code:
        1| orange = [[10, 100, 100], [18, 255, 255]]
        2| green = [[45, 100, 100], [75, 255, 255]]
        3|
        4| orange_and_green = MultiColor([orange, green])

    There are multiple built in "battery included" pre-made color object
    for instant use in testing and tuning
    List of colors:
      Red (MultiColorObject) : Red (low) + Red (high)
      Red (Low): [0, 100, 100], [8, 255, 255]
      Red (High): [172, 100, 100], [179, 255, 255]
      Note: in order to find red, use both ranges (low and high) and use the some of both results.
      Blue: [105, 100, 100], [135, 255, 255]
      Green: [45, 100, 100], [75, 255, 255]
      Yellow: [20, 100, 100], [55, 255, 255]
      Orange: [10, 100, 100], [18, 255, 255]
      Grey: [0, 0, 0], [179, 50, 195]
      Black: [0, 0, 0], [179, 255, 30]
      White: [0, 0, 200], [179, 20, 255]
      Teal: [110, 100, 100], [130, 255, 255]
      Purple: [135, 100, 100], [165, 255, 255]
    """

    def __init__(self, colors):
        """
        :param colors: A list of Color objects or tuple that can be turned into a color object
        Example:
              MultiColor([Color([15, 50, 50], [45, 255, 125]),
                         [[105, 175, 25], [115, 255, 255]]])
         These are different ways to create ranges of colors
         """
        self.colors = colors
        for index, color in enumerate(colors):
            if type(color) in (list, tuple):
                self.colors[index] = Color(*color)

    def convert(self, image: np.ndarray):
        """
        Thresholds an image using the color ranges,
        all pixels that are not in the color ranges defined are set to black (0, 0, 0)
        and all others to white (255, 255, 255)
        """
        if self.colors is None:
            raise ValueError("Cannot convert an image to a binary, no colors given.")
        if len(self.colors) == 0:
            raise ValueError("Cannot convert an image to binary, no colors given.")
        hsv_image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        binary_image = self.colors[0].threshold(hsv_image)
        if len(self.colors) < 1:
            return binary_image
        for color in self.colors[1:]:
            binary_image = cv2.bitwise_or(binary_image, color.threshold(hsv_image))

        return binary_image

    def validate(self, *args, **kwargs) -> bool:
        for color in self.colors:
            if type(color) not in Color:
                if not color.validate():
                    return False
        return True

    def __str__(self):
        return ', '.join([str(color) for color in self.colors])

    def __repr__(self):
        return 'MultiColor(0)'.format([repr(i) for i in self.colors])



import functools
import warnings

from .reverse_partial import ReversePartial


def keyword_partial(target_function):
    """
    A Decorator used to load other parameters to a function before applying it on a given input data
    This decorator is used to load parameters to the filter function, image processing filters, morphological functions
    sorters and other functions that act similarly.

    Preloading a function consists of pass all arguments except the first,
    which is then passed when the function is called.

    Example:

    For a given function:

        @keyword_partial
        def area_filter(contours, min_area, max_area):
            output_list = []
            ratio_list = []
            if type(contour_list) is not list:
                contour_list = [contour_list]
            for current_contour in contour_list:
                if min_area <= cv2.contourArea(current_contour) <= max_area:
                    output_list.append(current_contour)
                    ratio_list.append(current_contour)
            return output_list, ratio_list


    Instead of calling the function like other functions:

        A1| area_filter(list_of_contours, min_area=200, max_area=5000)

    The function needs to be called as follows:

        B1| activator = area_filter(min_area=200, max_area=5000)
        B2|
        B3| final_value = activator(list_of_contours)


    Vision objects use functions that are decorated with keyword_partial (contour_filter, image_filter,
    you can just pass the result of line B1 to the Vision object
    like so:

        C1| contour_filters = [some_filter(parameter1=5, parameter2=3), ovl.circle_filter(min_area_ratio=0.75)]
        C2|
        C3| vision = Vision(...,  contours_filters=contour_filters, ...)

    :param target_function: the function to be preloaded
    :return: a function (argument loader) that preloads (passes only some of the arguments)
             the wrapped function (target_function)
    """
    def argument_loader(*args, **kwargs):
        if args != ():
            warning_message = ("When passing parameters it is recommended to pass everything as keywords "
                               "in order to make it clear what parameters are passed."
                               "(Do: {0}(parameter1=value2, parameter2=value2) not {0}(value, value2))"
                               .format(target_function.__name__))
            warnings.warn(warning_message, SyntaxWarning)
        partial_function = ReversePartial(target_function, *args, **kwargs)
        return functools.update_wrapper(partial_function, target_function)

    wrapped_argument_loader = functools.update_wrapper(argument_loader, target_function)
    return wrapped_argument_loader




from typing import Callable, List, Any
import math

from .direction_monitors.direction_monitor import *
from ..camera_.camera_settings import *


class Director:
    """
    An object responsible of extrapolating directions from the detection of a Vision object
    This is the final step before sending/using the data.
    A director is built from a directing function and directing monitor
    """
    def __init__(self, directing_function: Callable, failed_detection: Any, target_amount: int = 1,
                 direction_monitors: List[DirectionMonitor] = None):
        """

        :param directing_function: the function that performs the initial direction calculation
        :param failed_detection: The value returned on a failed detection
        :param target_amount: the target amount to find, if the value is 0
                              the target amount (minimum amount) removed and is infinite
        :param direction_monitors: list of DirectionMonitor objects that alter the directions to be sent
        """
        self.directions = directing_function
        if not isinstance(target_amount, int):
            raise ValueError("Target amount must be an integer, got {} of type {}.".format(target_amount,
                                                                                           type(target_amount)))
        self.target_amount = target_amount if target_amount != 0 else math.inf
        self.direction_monitors = direction_monitors
        self.failed_detection = failed_detection

    def direct(self, contours: List[np.ndarray], image: np.ndarray, camera_settings, sorter=None) -> Any:
        """
        Returns the director for the given contours
        :param contours: the list of contours (numpy ndarrays) from which to extrapolate target direction
        :param image: the image from which the contours were taken
        :param camera_settings: The Camera settings, including position and calibration
        :param sorter: A sorter function to be applied on the contours, or None to not apply one
        :return: Depends on the directing function and the direction monitors,
                 usually a number or a string
        """
        if not len(contours) >= self.target_amount:
            directions = self.failed_detection
        else:
            if sorter is not None:
                contours = sorter(contours)
            if math.isfinite(self.target_amount):
                contours = contours[:self.target_amount]
            directions = self.directions(contours, image)

        return self.apply_direction_monitors(directions, contours, image, camera_settings)

    def apply_direction_monitors(self, directions: Any, contours: List[np.ndarray],
                                 image: np.ndarray, camera_settings: CameraSettings) -> Any:
        """
        Applies the list of direction monitors one after the other
        :param directions: the raw directions returned from the directions function
        :param contours: the list of contours that passed all filters
        :param image: the image from which the contours were found
        :param camera_settings: settings related to the camera, like position or calibration
        :return: the final direction result
        """
        if self.direction_monitors:
            for direction_monitor in self.direction_monitors:
                directions = direction_monitor.monitor(directions, contours, image, camera_settings)
                if direction_monitor.priority:
                    return directions
        return directions



from .network_location import NetworkLocation


class Connection:
    def send(self, data, *args, **kwargs):
        pass

    def receive(self, *args, **kwargs):
        pass

    def receive_from_location(self, network_location: NetworkLocation, *args, **kwargs):
        return self.receive(*args, **network_location, **kwargs,)

    def send_to_location(self, data, network_location: NetworkLocation, *args, **kwargs):
        return self.receive(*args, **network_location, **kwargs, data=data)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def close(self):
        pass



import cv2

from ..image_filters_.kernels import rectangle_kernel, is_odd_size
from ..image_filters_.image_filter import image_filter


@image_filter
def erosion(mask, kernel=None, iterations=None, destination=None,
            anchor=None, border_type=None, border_values=None):
    """
     a copy of cv2.erode with default kernel of 5 by 5
    (a logical operation on the binary mask,
    whether every pixel's value should stay as it is based on neighboring pixels,
    which neighbors are chosen by the kernel and its dimensions)
    Erode demands all chosen neighbors must be True (white)
    For more information:
    https://docs.opencv.org/3.0-beta/doc/py_tutorials/py_imgproc/py_morphological_ops/py_morphological_ops.html

    :param mask: the binary image where the erosion morphological function should be applied
    :param kernel: the kernel that should be used
    :param iterations: Number of times the function should be applied
    :param destination: where the new image should be saved
    :param anchor: position of the anchor within the element
    :param border_values: border value in case of a constant border
    :param border_type: Pixel extrapolation technique for the border of the image
    see: https://docs.opencv.org/3.4.2/d2/de8/group__core__array.html#ga209f2f4869e304c82d07739337eae7c5
    :return: the eroded binary mask
    """
    kernel = kernel or rectangle_kernel((5, 5))
    if not is_odd_size(kernel):
        raise ValueError("Received an invalid kernel: {} of type {}".format(kernel, type(kernel)))

    return cv2.erode(mask,
                     kernel,
                     dst=destination,
                     anchor=anchor,
                     iterations=iterations,
                     borderValue=border_values,
                     borderType=border_type)


@image_filter
def dilation(mask, kernel=None, iterations=None, destination=None,
             anchor=None, border_type=None, border_values=None):
    """
     a copy of cv2.dilate with default kernel of 5 by 5
    (a logical operation on the binary mask,
    whether every pixel's value should stay as it is based on neighboring pixels,
    which neighbors are chosen by the kernel and its dimensions)
    Dilation demands at least one chosen neighbors must be True (white)
    For more information:
    https://docs.opencv.org/3.0-beta/doc/py_tutorials/py_imgproc/py_morphological_ops/py_morphological_ops.html

    :param mask: the binary image where the erosion morphological function should be applied
    :param kernel: the kernel that should be used
    :param iterations: Number of times the function should be applied
    :param destination: where the new image should be saved
    :param anchor: position of the anchor within the element
    :param border_values: border value in case of a constant border
    :param border_type: sets the technique the pixels, that exceed the boundaries of the image for the use
                        of the kernel, are determined with (if they should mirror, copy
    see: https://docs.opencv.org/3.4.2/d2/de8/group__core__array.html#ga209f2f4869e304c82d07739337eae7c5
    :return: the dilated binary mask
    """
    kernel = kernel or rectangle_kernel((5, 5))
    if not is_odd_size(kernel):
        raise ValueError("Received an invalid kernel: {} of type {}".format(kernel, type(kernel)))
    return cv2.dilate(mask,
                      kernel,
                      dst=destination,
                      anchor=anchor,
                      iterations=iterations,
                      borderValue=border_values,
                      borderType=border_type)



import copy

from .display_image import *


def display_contours(image, contours, amount=0, delay=None, color=None, save_path=None):
    """
     Displays the image with the current_vision list of contours
    :param image: image from which the contours were taken from, numpy array or image path
    :param contours: the list of contours to display_image
    :param color: the color of the of the contours outline
    :param delay: delay for the wait key function, None means wait indefinitely,
                  add a delay when using this function in loops
    :param amount: amount of contours to display_image
    :param save_path: if the image should be saved, pass the wanted result path
    :return: the image with the drawn contours.
    """
    if type(image) is str:
        image = cv2.imread(image)
    image_for_display = copy.copy(image)
    contours = contours[0:amount] if amount > 0 else contours
    if color is None:
        color = (0, 0, 0)
    cv2.drawContours(image_for_display, contours, -1, color, 2)
    if type(save_path) is str:
        cv2.imwrite(save_path, image_for_display)
    display_image(image_for_display, delay=delay)
    return image_for_display



# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
from numpy import array, uint64, uint8, zeros
from random import randint, choice, uniform
import math
import cv2

from ..math_ import geometry
from ..thresholds.color_.color import Color
from ..thresholds.color_.multi_color import MultiColor

GRADE_N = 0

GRADE_NW = 1

GRADE_W = 2

GRADE_SW = 3

GRADE_S = 4

GRADE_SE = 5

GRADE_E = 6

GRADE_NE = 7

GRADE_CENTER = 8

list_of_grades = [GRADE_N, GRADE_NW, GRADE_W, GRADE_SW, GRADE_S, GRADE_SE, GRADE_E, GRADE_NE]


class Shape(object):
    def __init__(self, drawing_function, amount, dimensions_ranges, color, gradient=(0.25, 4)):
        self.drawing_function = drawing_function
        self.amount = amount
        self.color = color
        self.tone = None
        self.dimensions = dimensions_ranges
        self.gradient = gradient

    def draw(self, image, image_width, image_height):
        for i in range(self.amount):
            if type(self.drawing_function) is list:
                drawing_function = choice(self.drawing_function)
            else:
                drawing_function = self.drawing_function
            point, dimensions = self.__random_dimensions(image_width, image_height)
            self.tone = self.__random_color()
            drawing_function(image, point, dimensions, self.tone, self.__random_gradient())

    def __random_color(self):
        def random_range(low, high):
            return [randint(low[0], high[0]), randint(low[1], high[1]), randint(low[2], high[2])]

        color_range = self.color
        cr_type = type(color_range)
        if cr_type is Color:
            return random_range(color_range.low, color_range.high)
        elif cr_type is MultiColor:
            h, s, v = [], [], []
            for color in color_range.colors:
                j = random_range(color.low, color.high)
                h.append(j[0])
                s.append(j[1])
                v.append(j[1])
            return [choice(h), choice(s), choice(v)]

    def __random_dimensions(self, width, height):
        if type(self.dimensions) not in (list, tuple):
            raise TypeError('Dimensions is not a list or tuple!')
        if type(self.dimensions[0]) not in (int, list, tuple):
            raise TypeError('Dimensions must be either a range by itself or a list of ranges!')
        list_type = type(self.dimensions[0])
        for obj in self.dimensions:
            if type(obj) is not list_type:
                raise TypeError('Dimensions must be a range or a list of ranges, not both')

        if list_type == int:
            size = randint(self.dimensions[0], self.dimensions[1])
            return (randint(size, width - size), randint(0 + size, height - size)), size
        if list_type == list:
            dimensions = []
            for dimension in self.dimensions:
                size = randint(dimension[0], dimension[1])
                dimensions.append(size)
            size = max(dimensions)
            point = (randint(size, width - size), randint(size, height - size))
            return point, dimensions

    def __random_gradient(self, color_space='BGR'):
        if self.tone is None:
            return None
        if color_space in ['BGR', 'RGB']:
            gradient = []
            for i in self.tone:
                multi = 1
                if i > 127:
                    multi = -1
                gradient.append(multi)
            for idx, tone in enumerate(self.tone):
                if tone == 0:
                    self.tone[idx] = 1
            r1, r2 = self.tone[1] / self.tone[0], self.tone[2] / self.tone[0]
            r0 = round(uniform(gradient[0], gradient[1]), 3)
            r1 *= r0 * gradient[1]
            r2 *= r0 * gradient[2]
            r0 += gradient[0]
            b = str(randint(-5, 5))
            if color_space is 'RGB':
                return str(r2) + '* x + ' + b, str(r1) + '* x + ' + b, str(r0) + '* x + ' + b
            return str(r0) + '* x + ' + b, str(r1) + '* x + ' + b, str(r2) + '* x + ' + b


def gradient_circle(image, center, radius, color=(0, 0, 0), gradient=('2*x', '2*x', '2*x'), tangent_angle=0):
    """
     Draws a circle with a gradient according to the given traits
    :param image: The that the circle should be drawn on, an already open image as a numpy array
                 (opened with cv2.imread)
    :param center: the center of the circle a tuple in the form of (x,y), in pixels, (0,0) is the top left corner
    :param radius: the radius of the circle to be drawn, in pixels
    :param color: the starting color of the circle, in the color space of the image
    :param gradient: the change in color over x and y, in the form of linear equations
    :param tangent_angle: the angle of the compared to the center of the circle  where the gradient starts
    :return: None
    """
    if tangent_angle == 0:
        tangent_angle = randint(0, 360)
    ce = geometry.calculate_math_expression
    for i in range(radius):
        current_color = (color[0] + int(ce(gradient[0], radius - i)),
                         color[1] + int(ce(gradient[1], radius - i)),
                         color[2] + int(ce(gradient[2], radius - i)))
        circle_center = (int(center[0] + (radius - i) * math.sin(math.radians(tangent_angle))),  # x value
                         int(center[1] + (radius - i) * math.cos(math.radians(tangent_angle))))  # y value
        cv2.circle(image, circle_center, i, current_color, 2)


def center_gradient_circle(image, center, radius, color=(0, 0, 0), gradient=('2*x', '2*x', '2*x')):
    """
     Draws a circle with a gradient according to the given traits
    :param image: The that the circle should be drawn on, an already open image as a numpy array
    :param center: the center of the circle a tuple in the form of (x,y), in pixels, (0,0) is the top left corner
    :param radius: the radius of the circle to be drawn, in pixels
    :param color: the starting color of the circle, in the color space of the image
    :param gradient: the change in color over x and y, in the form of linear equations
    :return: None
    """
    if type(image) == str:
        image = cv2.imread(image)

    ce = geometry.calculate_math_expression
    for i in range(radius):
        current_color = (color[0] + int(ce(gradient[0], radius - i)),
                         color[1] + int(ce(gradient[1], radius - i)),
                         color[2] + int(ce(gradient[2], radius - i)))
        cv2.circle(image, center, i, current_color, 2)


def rotated_rectangle(image, start_point, dimensions, angle=360, color=(0, 0, 0), thickness=1):
    """
     Draws a rotated rectangle
    :param image: The image on which the rectangle should be drawn, a numpy array (an image opened with imread)
    :param start_point: The top left corner, used also as the rotation axis of the rectangle
                        (0, 0) is the top left corner
    :param dimensions: width, height of the rectangle in pixels
    :param color: color of the rectangle in the image's color space
    :param thickness: the thickness of the outline of the rectangle in pixels, -1 means filled
    :param angle: the angle of rotation in degrees
    :return: angle of rotation
    """
    if type(image) == str:
        image = cv2.imread(image)
    w, h = dimensions[0], dimensions[1]
    x, y = start_point
    a = [x, y]
    b, d = [x + int(math.cos(math.radians(angle)) * w), y - int(math.sin(math.radians(angle)) * w)], \
           [x + int(math.sin(math.radians(angle)) * h), y + int(math.cos(math.radians(angle)) * h)]
    vd = [d[0] - x, d[1] - y]
    vb = [b[0] - x, b[1] - y]
    c = [a[0] + vd[0] + vb[0], a[1] + vd[1] + vb[1]]
    pts = array([a, b, c, d], dtype=uint64)
    pts = pts.reshape((-1, 1, 2))
    cv2.polylines(image, [pts], True, color=color, thickness=thickness)
    return angle


def gradient_rectangle(image, start_point, dimensions, color=(0, 0, 0), grade=-1, angle=360.0,
                       gradient=('2 * x', '2 * x', '2 * x')):
    """
     Draws a rectangle with a changing color
    :param image: image to draw on
    :param start_point: the top left corner of the rectangle from which it is drawn
    :param dimensions: width, height of the rectangle
    :param color: the starting color of the gradient
    :param grade: the starting location of the gradient
    :param angle: optional parameter, the angle of rotation of the rectangle
    :param gradient: the grade of change of the color, it is a tuple of 3 polynomial equations
    (they can be unsimplified) written in python like code as  strings.
    The color space used affects the what each index changes (hsv, bgr, rgb, lab etc)
    the grade according x value is calculated as part of the given equation and then added to the starting value
    so the final value for the first value (regardless of color space) is color[0] + f(x) where f(x) is the result
    of the equation given at the coordinate x.
    Example:
    -----------
    if in the rgb color space:
    ('2 * x', '-3 * x', '2.5 * x')
    increases the red value by 2 for every pixel decreases the green
     value by 3 for every pixel and the blue value by 2.5
    (it is rounded), make sure when using negative values that the corresponding
    starting value is large enough or it will stale at black (or white)
    if in the hsv color space:
    ('0.25 * (x**2) + 2', '(x + 3) / -5', 'x + 3')
    the change in the h value increases by its value squared then multiplied by 0.25 and finally increased by 2
    thus the final h value for every pixel is: starting h value (color[0]) + the result of the
    the s value increases is increased by 3 then divided by -5
    the v value is starting color + 3 + number of pixel (increases by 1 per pixel)
    :return: None
    """
    width, height = dimensions[0], dimensions[1]
    if angle == 360:
        angle = round(uniform(0, 359), 2)
    if grade == -1:
        grade = choice(list_of_grades)

    def draw_straight(point, code):
        x_val, y_val = point[0], point[1]
        ce = geometry.calculate_math_expression
        ratio = width / height
        if code in [GRADE_N, GRADE_S]:
            for i in range(0, height):
                if code is GRADE_N:
                    current_color = (color[0] + int(ce(gradient[0], i)),
                                     color[1] + int(ce(gradient[1], i)),
                                     color[2] + int(ce(gradient[2], i)))
                else:
                    current_color = (color[0] + int(ce(gradient[0], height - i)),
                                     color[1] + int(ce(gradient[1], height - i)),
                                     color[2] + int(ce(gradient[2], height - i)))
                cv2.line(image, (x_val, y_val + i), (x_val + width, y_val + i), color=current_color)
        elif code in [GRADE_W, GRADE_E]:
            for i in range(0, width):
                if code is GRADE_W:
                    current_color = (color[0] + int(ce(gradient[0], i)),
                                     color[1] + int(ce(gradient[1], i)),
                                     color[2] + int(ce(gradient[2], i)))
                else:
                    current_color = (color[0] + int(ce(gradient[0], width - i)),
                                     color[1] + int(ce(gradient[1], width - i)),
                                     color[2] + int(ce(gradient[2], width - i)))
                cv2.line(image, (x_val + i, y_val), (x_val + i, y_val + height), color=current_color)
        elif code is GRADE_NW:

            for i in range(0, height):
                c_width = int(width - (i * ratio))
                c_height = height - i
                current_color = (color[0] + int(ce(gradient[0], height - i)),
                                 color[1] + int(ce(gradient[1], height - i)),
                                 color[2] + int(ce(gradient[2], height - i)))
                cv2.rectangle(image, start_point, (x_val + c_width, y_val + c_height), current_color, -1)
        elif code is GRADE_NE:

            for i in range(0, height):
                c_width = int(width - (i * ratio))
                c_height = height - i
                new_point = (start_point[0] + width, start_point[1])
                x_val, y_val = new_point[0], new_point[1]
                current_color = (color[0] + int(ce(gradient[0], height - i)),
                                 color[1] + int(ce(gradient[1], height - i)),
                                 color[2] + int(ce(gradient[2], height - i)))
                cv2.rectangle(image, new_point, (x_val - c_width, y_val + c_height), current_color, -1)
        elif code is GRADE_SW:

            for i in range(0, height):
                c_width = width - int(i * ratio)
                c_height = height - i
                new_point = (start_point[0], start_point[1] + height)
                x_val, y_val = new_point[0], new_point[1]
                current_color = (color[0] + int(ce(gradient[0], height - i)),
                                 color[1] + int(ce(gradient[1], height - i)),
                                 color[2] + int(ce(gradient[2], height - i)))
                cv2.rectangle(image, new_point, (x_val + c_width, y_val - c_height), current_color, -1)
        elif code is GRADE_SE:

            for i in range(0, height):
                c_width = width - int(i * ratio)
                c_height = height - i
                new_point = (start_point[0] + width, start_point[1] + height)
                x_val, y_val = new_point[0], new_point[1]
                current_color = (color[0] + int(ce(gradient[0], height - i)),
                                 color[1] + int(ce(gradient[1], height - i)),
                                 color[2] + int(ce(gradient[2], height - i)))
                cv2.rectangle(image, new_point, (x_val - c_width, y_val - c_height), current_color, -1)

    def draw_rotated(point, code):
        ce = geometry.calculate_math_expression
        ratio = width / height
        if code in [GRADE_N, GRADE_S]:
            for i in range(0, height):
                if code is GRADE_N:
                    current_color = (color[0] + int(ce(gradient[0], i)),
                                     color[1] + int(ce(gradient[1], i)),
                                     color[2] + int(ce(gradient[2], i)))
                else:
                    current_color = (color[0] + int(ce(gradient[0], height - i)),
                                     color[1] + int(ce(gradient[1], height - i)),
                                     color[2] + int(ce(gradient[2], height - i)))
                rotated_rectangle(image, point, (width, height - i), color=current_color, thickness=3, angle=angle)
        elif code in [GRADE_W, GRADE_E]:
            for i in range(0, width):
                if code is GRADE_W:
                    current_color = (color[0] + int(ce(gradient[0], i)),
                                     color[1] + int(ce(gradient[1], i)),
                                     color[2] + int(ce(gradient[2], i)))
                else:
                    current_color = (color[0] + int(ce(gradient[0], width - i)),
                                     color[1] + int(ce(gradient[1], width - i)),
                                     color[2] + int(ce(gradient[2], width - i)))
                rotated_rectangle(image, point, (width - i, height), color=current_color, thickness=3, angle=angle)
        elif code is GRADE_NW:
            for i in range(0, height):
                c_width = int(width - (i * ratio))
                c_height = height - i
                current_color = (color[0] + int(ce(gradient[0], height - i)),
                                 color[1] + int(ce(gradient[1], height - i)),
                                 color[2] + int(ce(gradient[2], height - i)))
                rotated_rectangle(image,
                                  start_point,
                                  (c_width, c_height),
                                  color=current_color,
                                  thickness=3,
                                  angle=angle)
        elif code is GRADE_NE:
            for i in range(0, height):
                c_width = int(width - (i * ratio))
                c_height = height - i
                new_point = (start_point[0] + width, start_point[1])
                current_color = (color[0] + int(ce(gradient[0], height - i)),
                                 color[1] + int(ce(gradient[1], height - i)),
                                 color[2] + int(ce(gradient[2], height - i)))
                rotated_rectangle(image,
                                  new_point,
                                  (- c_width, c_height),
                                  color=current_color,
                                  thickness=3,
                                  angle=angle)
        elif code is GRADE_SW:
            for i in range(0, height):
                c_width = width - int(i * ratio)
                c_height = height - i
                new_point = (start_point[0], start_point[1] + height)
                current_color = (color[0] + int(ce(gradient[0], height - i)),
                                 color[1] + int(ce(gradient[1], height - i)),
                                 color[2] + int(ce(gradient[2], height - i)))
                rotated_rectangle(image,
                                  new_point,
                                  (c_width, - c_height),
                                  color=current_color,
                                  thickness=3,
                                  angle=angle)
        elif code is GRADE_SE:
            for i in range(0, height):
                c_width = width - int(i * ratio)
                c_height = height - i
                new_point = (start_point[0] + width, start_point[1] + height)
                current_color = (color[0] + int(ce(gradient[0], height - i)),
                                 color[1] + int(ce(gradient[1], height - i)),
                                 color[2] + int(ce(gradient[2], height - i)))
                rotated_rectangle(image,
                                  new_point,
                                  (- c_width, - c_height),
                                  color=current_color,
                                  thickness=3,
                                  angle=angle)

    if angle == 0:
        draw_straight(start_point, grade)
    else:
        draw_rotated(start_point, grade)


def create_color_image(color, width=320, height=240, color_space=None):
    """
     Creates an image that is completely black
    :param color: The color of the image, in the BGR Color Space
    :param width: Width of the image in pixels, default is 320 pixels
    :param height: Height of the image in pixels, default is 240 pixels
    :param color_space: The color space of the
    :return: The image
    :rtype: numpy array
    """
    image = zeros((height, width, 3), dtype=uint8)
    image[:, :] = color
    if color_space is not None:
        return cv2.cvtColor(image, color_space)
    return image


class GenerateImage:
    """
    An image generating object, can create image with given shapes
    """

    def __init__(self, width, height, image_color, true_shape, false_shapes=()):
        """
         Generates random images
        :param width: width of the image in pixels
        :param height: height of the image in pixels
        :param true_shape: the shape object of the true_shape target
        :param false_shapes: a list of the false_shapes shapes
        :param image_color: the image background color
        """
        self.true_shape = true_shape
        self.false_shapes = false_shapes
        self.width, self.height = width, height
        self.image = None
        self.image_color = image_color

    def __random_color(self):
        def random_range(low, high):
            return [randint(low[0], high[0]), randint(low[1], high[1]), randint(low[2], high[2])]

        color_range = self.image_color
        cr_type = type(color_range)
        if cr_type is Color:
            return random_range(color_range.low, color_range.high)
        elif cr_type is MultiColor:
            hue, saturation, brightness = [], [], []
            for ranges in color_range.ranges:
                j = random_range(ranges.low, color_range.high)
                hue.append(j[0])
                saturation.append(j[1])
                brightness.append(j[1])
            return [choice(hue), choice(saturation), choice(brightness)]

    def generate_image(self):
        self.image = create_color_image(self.__random_color(), self.width, self.height)
        for f in self.false_shapes:
            f.draw(self.image, self.width, self.height)
        self.true_shape.draw(self.image, self.width, self.height)
        return self.image


def draw_enclosing_circle(image, contour, color=(0, 0, 0), thickness=1, display_data=True):
    """
     Draws the enclosing (bounding) circle on the given contour
    :param image: image where the contour is, a numpy array or file path
    :param contour: The numpy array describing the shape that should be enclosed by a circle
    :param color: the color of the circle
    :param thickness: the thickness in pixels of the circle
    :param display_data:a bool that determines if the radius of the circle should be displayed
    :return: The image
    :rtype: a numpy array (ndarray)
    """

    def ensure(image_p, rad, cen, inc):
        w, h, _ = image_p.shape
        o1 = int(cen[1] - rad - 1.5 * inc)
        o2 = int(cen[1] + rad + 1.5 * inc)
        o3 = int(cen[0] - rad - 1.5 * inc)
        o4 = int(cen[0] + rad + 1.5 * inc)
        if o1 >= 0:
            return cen[0], o1
        elif o2 <= h:
            return cen[0], o2
        elif o3 >= 0:
            return o3, cen[1]
        elif o4 <= w:
            return o4, cen[1]

    if type(image) is str:
        image = cv2.imread(image)
        if image is None:
            raise ValueError('Given image is an invalid path (check if the image path exists)'
                             ' and image could not be opened.')
    center, radius = cv2.minEnclosingCircle(contour)
    c = (int(center[0]), int(center[1]))
    r = int(radius)
    cv2.circle(image, c, r, color, int(thickness))
    if display_data:
        font = cv2.FONT_HERSHEY_SCRIPT_SIMPLEX
        c = ensure(image, r, c, 5)
        cv2.putText(image, str(r), c, font, 0.5, color, 2)
    return image


def draw_bounding_rect(image, contour, color=(0, 0, 0), thickness=1, display_data=True):
    """
     draws the straight bounding rectangle around the given contour.
    :param image: Image where the contour is, a numpy array or file path to the image
    :param contour: The numpy array describing the shape that should be bounded by a rectangle
    :param color: color of the outline of the bounding rectangle, bgr color space
    :param thickness: thickness of the bounding rectangle in pixels, -1 fills the shape
    :param display_data: a bool that determines if the
    :return: the image
    :rtype: a numpy array
    """
    if type(image) is str:
        image = cv2.imread(image)
        if image is None:
            raise ValueError('Given image is an invalid path (check if the image path exists)'
                             ' and image could not be opened.')

    x, y, w, h = cv2.boundingRect(contour)
    cv2.rectangle(image, (x, y), (x + w, y + h), color=color, thickness=thickness)
    if display_data:
        font = cv2.FONT_HERSHEY_SCRIPT_SIMPLEX
        text = 'w=' + str(w) + ', h=' + str(h)
        cv2.putText(image, text, (x, y - 5), font, 0.75, color, 2)
    return image



# Copyright 2018 - 2019 Ori Ben-Moshe - All rights reserved.
from typing import Any, Union
from threading import Thread
import uuid
import platform
import numpy as np
import cv2


class Camera:
    def __init__(self, source: Union[str, int, cv2.VideoCapture] = 0,
                 image_width: int = 320, image_height: int = 240, start_immediately=True):
        """
        :param source: the source of the camera, can be a number, a device name or any valid source
                       for the cv2.VideoCapture object.
        :param image_width: the width of images to be captured in pixels
        :param image_height: the height of images to be captured in pixels
        """
        self.stream = cv2.VideoCapture(source)
        self.name = uuid.uuid4()

        if image_width:
            self.stream.set(3, image_width)
        if image_height:
            self.stream.set(4, image_height)
        self.grabbed, self.frame = self.stream.read()
        self.stopped = False
        self.camera_thread: Union[None, Thread] = None
        self.start_immediately = start_immediately
        if start_immediately:
            self.start()

    def set_exposure(self, exposure_value) -> None:
        """
         Sets the exposure for the camera.
        Value range depend on the camera some use the value as the exponential so negative values
        :param exposure_value:
        """
        return self.stream.set(15, exposure_value)

    def start(self) -> "Camera":
        """
        starts the image taking thread
        :return: itself
        """
        if self.stopped:
            return self
        self.stopped = False
        self.camera_thread = Thread(target=self.update, args=()).start()
        return self

    def update(self):
        """
        Takes a new image while not stopped
        :return:
        """
        while not self.stopped:
            self.grabbed, self.frame = self.stream.read()

    def read(self) -> [bool, np.ndarray]:
        """
         Returns the ret val and the frame
        :return: if the image was taken successfully, the image numpy
        :rtype: bool, numpy.array
        """
        return self.grabbed, self.frame

    def stop(self) -> None:
        """
        Stops the camera thread.
        """
        self.stopped = True

    def release(self) -> None:
        """
         stops the camera thread and releases the camera
        :return: None
        """
        self.stopped = True
        self.stream.release()

    def set(self, property_id: int, value) -> None:
        """
         sets the value of the given property to the given value
         :param property_id: the property number, for more information:
                        https://docs.opencv.org/3.1.0/d8/dfe/classcv_1_1VideoCapture.html
        :param value: the value to be set, a number
        :return: None
        """
        self.stream.set(property_id, value)

    def get(self, property_id) -> Any:
        """
         retrieves the value of a property based on its id
        :param property_id: the property number, for more information:
                        https://docs.opencv.org/3.1.0/d8/dfe/classcv_1_1VideoCapture.html
        :return: the value
        """
        if isinstance(property_id, int):
            raise ValueError("Given Property id {} was not valid, please refer to the documentation: "
                             "https://docs.opencv.org/3.1.0/d8/dfe/classcv_1_1VideoCapture.html".format(property_id))
        if not 0 <= type(property_id) < 19:
            raise ValueError("Invalid property id ({}), please refer to the documentation".format(property_id))
        return self.stream.get(property_id)

    def isOpened(self) -> bool:
        """
        Returns true_shape if the camera is opened and false_shapes otherwise
        :return: if the camera is open
        """
        return self.stream.isOpened()

    def getBackendName(self) -> str:
        """
        :return: returns the backend name of a camera
        :raises: SystemError on windows
        """
        if platform.system() == "Windows":
            raise SystemError("This method only available on Linux Operating Systems.")
        return self.stream.getBackendName()



# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
import numpy as np
import copy
import cv2
from typing import *

from ..threshold import Threshold

BaseForColor = NewType("BaseForColor", Union[int, Tuple[Tuple[int, int, int], Tuple[int, int, int]]])
SERIALIZED_COLOR_KEYS = {"high", "low"}


def validate(value, ceiling):
    """
     Checks if val is positive and less than the ceiling value.
    :param value: The value to be checked (usually an int)
    :param ceiling: The highest value "val" can be. (usually an int)
    :return: True if val is less than ceiling and not negative
    """
    value = int(value)
    return 0 <= value < ceiling


def assert_hsv(hsv_point):
    """
     Makes sure the hsv point(or vector) given in the parameter has valid values.
    :param hsv_point: a 3 length list with 3 ints describing a point the hsv color space
    that describe a color limit in a range for a findContour function.
    :return: True if valid False if not.
    :rtype: bool
    """

    return validate(hsv_point[0], 180) and validate(hsv_point[1], 256) and validate(hsv_point[2], 256)


class Color(Threshold):
    """
    Color is a Threshold object( an object that turns an image to a binary image)
    Color object thresholds using 2 HSV color ranges.
    Read more about HSV here:
    https://www.lifewire.com/what-is-hsv-in-design-1078068

    HSV Ranges in OVL (And in the underlying open-cv (cv2)) are:
    Hue - 0 to 179 (360/2 so divide any regular HSV value you use by 2)
        So if Yellow is 40 - 80 (degrees) in regular HSV palette its 20 to 40 in ovl
    Saturation - 0 to 255 (255 is equal to 1 or 100%)
    Value - 0 to 255 (255 is equal to 1 or 100%)

    2 ranges are passed to the Constructor:
        1) Low HSV Range - the lowest acceptable H S V values ([low_H, low_S, low_V])
        2) High HSV Range - the highest acceptable H S V values ([high_H, high_S, high_V])

    Example code:
        1| low_range = [15, 100, 100]
        2| high_range = [45, 255, 255]
        3| color = ovl.Color(low_range, high_range)

    The color object can the be passed to a Vision object to threshold binary images
    Threshold object can be used by themselves using the color.convert() method.

    NOTE: Threshold objects automatically convert images to HSV (From the default BGR)

    There are multiple built in "battery included" pre-made color object
    for instant use in testing and tuning
    List of colors:
      Red (MultiColorObject) : Red (low) + Red (high)
      Red (Low): [0, 100, 100], [8, 255, 255]
      Red (High): [172, 100, 100], [179, 255, 255]
      Note: in order to find red, use both ranges (low and high) and use the some of both results.
      Blue: [105, 100, 100], [135, 255, 255]
      Green: [45, 100, 100], [75, 255, 255]
      Yellow: [20, 100, 100], [55, 255, 255]
      Orange: [10, 100, 100], [18, 255, 255]
      Grey: [0, 0, 0], [179, 50, 195]
      Black: [0, 0, 0], [179, 255, 30]
      White: [0, 0, 200], [179, 20, 255]
      Teal: [110, 100, 100], [130, 255, 255]
      Purple: [135, 100, 100], [165, 255, 255]
    """
    def validate(self, *args, **kwargs):
        return assert_hsv(self.low_bound) and assert_hsv(self.high_bound)

    def __init__(self, low: BaseForColor, high: BaseForColor):
        """
         Constructor for the Color used to turn images to binary images based on
         HSV color space (pixels that are in the specified HSV Range)
         H - 0 - 179
         S - 0 - 255
         V - 0 - 255
        :param high: high hsv limit of the color
        :param low: low hsv limit of the color
        """

        if type(low) is tuple:
            low = list(low)
        if type(low) is int:
            low = [low, 100, 100]
        if type(high) is tuple:
            high = list(high)
        if type(high) is int:
            high = [high, 255, 255]
        self.__low_bound = np.array(low)
        self.__high_bound = np.array(high)

    def threshold(self, image: np.ndarray) -> np.ndarray:
        return cv2.inRange(image, self.low, self.high)

    def convert(self, image: np.ndarray) -> np.ndarray:
        """
         converts a given image to hsv and then thresholds and returns the binary mask
        :param image: a BGR image
        :return: binary mask
        :rtype: numpy array
        """
        hsv_image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        return self.threshold(hsv_image)

    def __repr__(self):
        return 'Color({}, {})'.format(repr(self.low_bound), repr(self.high_bound))

    def copy(self):
        """
        Duplicates the Color object so that changes do not affect the original color
        Useful for modifying BuiltInColors without changing the default
        :return: a copy of the color object
        """
        return copy.deepcopy(self)

    @property
    def low_bound(self):
        if isinstance(self.__low_bound, np.ndarray):
            return self.__low_bound.tolist()
        return self.__low_bound

    @property
    def high_bound(self):
        if isinstance(self.__high_bound, np.ndarray):
            return self.__high_bound.tolist()
        return self.__high_bound

    @property
    def low(self):
        """
         Returns the low hsv limit of the color.
        :return: An uint8 numpy array with the low limit of the color.
        :rtype: uint8 numpy array
        """
        return self.__low_bound

    @property
    def high(self):
        """
         Returns the high hsv limit of the color.
        :return: An uint8 numpy array with the high limit of the color.
        :rtype: uint8 numpy array
        """
        return self.__high_bound

    def __str__(self):
        return self.__repr__()



import cv2


class CameraCalibration:
    def __init__(self, object_points, image_points, image_dimensions, alpha=0, save_raw=False):
        self.image_dimensions = image_dimensions
        if save_raw:
            self.image_points = image_points
            self.object_points = object_points
        else:
            self.image_points = None
            self.object_points = None
        calibration = cv2.calibrateCamera(object_points, image_points, image_dimensions, None, None)
        (_, camera_matrix, distortion_coefficients,
         rotation_vectors, translation_vectors) = calibration
        optimal_matrix, region_of_image = cv2.getOptimalNewCameraMatrix(camera_matrix,
                                                                        distortion_coefficients,
                                                                        image_dimensions,
                                                                        alpha)
        self.optimal_matrix = optimal_matrix
        self.region_of_image = region_of_image
        self.camera_matrix = camera_matrix
        self.rotation_vectors = rotation_vectors
        self.translation_vectors = translation_vectors
        self.distortion_coefficients = distortion_coefficients

    def undistort_image(self, image):
        """
         Removes distortion created by imperfections in the camera.
        :param image: The image (numpy array) that should be undistorted
        :return: an undistorted copy of the image
        """
        return cv2.undistort(image, self.camera_matrix, self.distortion_coefficients, None, self.optimal_matrix)



from typing import Any


class NetworkLocation:
    """
    An object used to specify parameters for a Connection.send  and Connection.receive functions
    This is used to create a specific address / port for a connection and allows automated specification
    for a place to send / receive.

    A common use is for MultiVision to specify a "location" from which to read the current vision to be used.
    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def keys(self):
        return self.kwargs.keys()

    def __getitem__(self, key) -> Any:
        return self.kwargs[key]



import typing

from .connection import Connection


class SerialConnection(Connection):
    def __init__(self, port: typing.Union[int, str], baudrate: int = 9600, *args, **kwargs):
        from serial import Serial

        self.socket = Serial(port, baudrate, *args, **kwargs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.socket.close()

    def send(self, data, *args, **kwargs):
        self.socket.write(data)

    def receive(self, *args, **kwargs):
        return self.socket.read()



# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
from typing import List

from ..math_.geometry import *
from ..math_.geometry import distance_between_points as dbp, distance_from_frame as dff
from .contour_filter import contour_filter


@contour_filter
def image_center_filter(contour_list, image_dimensions=(320, 240), max_dist=0.7):
    """
    Filters out contours that their center is not close enough to the center of the image
    :param contour_list: a list of contours to be filtered
    :param image_dimensions: the size of the image(width, height)
    :param max_dist: the maximum percent of difference
    :return: list of contours within the maximum distance from the image center
    """
    output = []
    ratio = []
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    if type(contour_list) is not list:
        contour_list = [contour_list]

    for current_contour in contour_list:
        current_contour_center = contour_center(current_contour)
        distance_ratio = (dbp(current_contour_center, image_center)
                          / float(dff(current_contour_center, image_dimensions)))
        if distance_ratio >= max_dist:
            output.append(current_contour)
            ratio.append(distance_ratio)
    return output, ratio


@contour_filter
def distance_filter(contour_list, point, min_dist=0, max_dist=50):
    """
    Filters out contours that their center is not close enough
    to the given (x, y) point
    :param contour_list: a list of contours to be filtered
    :param point: the point from which the contours should be filtered a tuple (or list) of 2 numbers.
    :param max_dist: the maximum distance from the point in pixels
    :param min_dist: the minimum distance from the point in pixels
    :return: the filtered contour list
    """
    output = []
    ratio = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        current_contour_center = contour_center(current_contour)
        distance_from_center = dbp(current_contour_center, point)
        if max_dist >= distance_from_center >= min_dist:
            output.append(current_contour)
            ratio.append(distance_from_center)
    return output, ratio


@contour_filter
def absolute_distance_filter(contour_list, max_dist=50, min_dist=0, image_dimensions=(320, 240)):
    """
    Filters out contours that their center is not close enough
    to the center of the image
    :param contour_list: a list of contours to be filtered
    :param image_dimensions: the size of the image(width, height)
    :param max_dist: the maximum distance from the center in pixels
    :param min_dist: the minimum distance from the center in pixels
    :return:
    """
    output = []
    ratio = []
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        current_contour_center = contour_center(current_contour)
        distance_from_center = dbp(current_contour_center, image_center)
        if max_dist >= distance_from_center >= min_dist:
            output.append(current_contour)
            ratio.append(distance_from_center)
    return output, ratio


@contour_filter
def length_filter(contour_list, min_length=50, max_length=76800):
    """
    Receives a list of contours and removes ones that are not long enough
    Note: for "open" contours only!
    :param contour_list: list of contours (numpy array) to be filtered
    :param min_length: minimum length of a contour (in pixels)
    :param max_length: maximum length of a contour (in pixels)
    :return: list of filtered contours and list of the lengths
    """
    output = []
    ratio = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        perimeter = cv2.arcLength(current_contour, False)
        if min_length >= perimeter >= max_length:
            output.append(current_contour)
            ratio.append(perimeter)
    return output, ratio


@contour_filter
def straight_square_filter(contour_list, min_area_ratio=0.8, min_len_ratio=0.95):
    """
    Receives a list of contours and returns
    only the ones that are approximately square
    Relation checked is [minimum ratio < (Circle radius / width * height * (square root of 2)) < maximum ratio]
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_len_ratio: maximum ratio between radius and sides of bounding rectangle
    :param min_area_ratio: the minimum ratio between the area of the contour and the area of the bounding shape
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, contour_width, contour_height = rectangle_fill_ratio_straight(current_contour)
        peri = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and len(approximation) == 4:
            _, enclosing_radius = cv2.minEnclosingCircle(current_contour)
            bounding_rectangle_diagonal = (2 ** 0.5) * ((contour_width * contour_height) ** 2)
            enclosing_circle_radius = 2 * enclosing_radius
            radius_ratio = enclosing_circle_radius / bounding_rectangle_diagonal
            if radius_ratio >= min_len_ratio:
                output_list.append(current_contour)
                ratio_list.append(fill_ratio * radius_ratio)
    return output_list, ratio_list


@contour_filter
def rotated_square_filter(contour_list, min_area_ratio=0.8, min_ratio=0.95, max_ratio=1.05):
    """
    Receives a list of contours
    and returns only the ones that are approximately square
    Relation checked is [minimum ratio < (Circle radius / width * height * (square root of 2)) < maximum ratio]
    :param contour_list: List of Contours to filter
    :param max_ratio: maximum ratio between radius and sides of bounding rectangle
    :type max_ratio: float
    :param min_ratio: minimum ratio between radius and sides of bounding rectangle
    :type min_ratio: float
    :param min_area_ratio: minimum ratio between the area of the contours and the bounding shape
    :return: the contour list filtered.
     """
    ratio_list = []
    output_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, bounding_width, bounding_height = rotating_rectangle_fill_ratio(current_contour)
        _, enclosing_radius = cv2.minEnclosingCircle(current_contour)
        contour_perimeter = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * contour_perimeter, True)
        if fill_ratio > min_area_ratio and len(approximation) == 4:
            diagonal_length = (2 ** 0.5) * ((bounding_width * bounding_height) ** 0.5)
            enclosing_circle_diameter = 2 * enclosing_radius
            radius_ratio = enclosing_circle_diameter / diagonal_length
            if min_ratio < radius_ratio < max_ratio:
                output_list.append(current_contour)
                if radius_ratio > 1:
                    radius_ratio = 1 / radius_ratio
                ratio_list.append(fill_ratio * radius_ratio)
    return output_list, ratio_list


@contour_filter
def vertical_rectangle_filter(contour_list, min_area_ratio=0.):
    """
    Receives a list of contours
    and returns only the ones that are approximately a vertical rectangle
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_area_ratio: minimum ratio between the area of the contours and the bounding shape
    :return: the contour list filtered.
     """
    ratio_list = []
    output_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, contour_width, contour_height = rectangle_fill_ratio_straight(current_contour)
        peri = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and contour_width < contour_height and len(approximation) == 4:
            output_list.append(current_contour)
            ratio_list.append(fill_ratio)
    return output_list, ratio_list


@contour_filter
def horizontal_rectangle_filter(contour_list, min_area_ratio=0.8):
    """
    Receives a list of contours and returns only those that are approximately a horizontal rectangle
    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_area_ratio
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, contour_width, contour_height = rectangle_fill_ratio_straight(current_contour)
        perimeter = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * perimeter, True)
        if fill_ratio > min_area_ratio and contour_width > contour_height and len(approximation) == 4:
            output_list.append(current_contour)
            ratio_list.append(fill_ratio)
    return output_list, ratio_list


@contour_filter
def area_filter(contour_list, min_area=200, max_area=math.inf):
    """
    Filters contours that are not within the threshold of area (in pixels)
    :param max_area: maximum area of a contour (Inclusive) default is no limit (infinity)
    :param min_area: minimum area of a contour (Inclusive) set to 0 for no lower limit.
    :param contour_list: List of Contours to filter
    :return: the contour list filtered.
    """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        if min_area <= cv2.contourArea(current_contour) <= max_area:
            output_list.append(current_contour)
            ratio_list.append(current_contour)
    return output_list, ratio_list


@contour_filter
def straight_rectangle_filter(contour_list, min_area_ratio=0.8):
    """
     receives a list of contours and returns only those that are approximately a rectangle
    and that are not rotated

    :param contour_list: List of Contours to filter
    :type contour_list: List
    :param min_area_ratio: The minimum ratio between the rectangle and the contour
    :type min_area_ratio: float
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, _, _ = rectangle_fill_ratio_straight(current_contour)
        perimeter = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * perimeter, True)
        if fill_ratio > min_area_ratio and len(approximation) == 4:
            output_list.append(current_contour)
            ratio_list.append(fill_ratio)
    return output_list, ratio_list


@contour_filter
def rotated_rectangle_filter(contour_list, min_area_ratio=0.8):
    """
    Receives a list of contours and returns only those that are approximately a rectangle regardless
    of the angle of rotation.
    :param contour_list: List of Contours to filter
    :param min_area_ratio: the minimum ratio between the contour area and the bounding shape
    :return: the contour list filtered.
     """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, w, h = rotating_rectangle_fill_ratio(current_contour)
        peri = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, 0.02 * peri, True)
        if fill_ratio > min_area_ratio and len(approximation) == 4:
            ratio_list.append(fill_ratio)
            output_list.append(current_contour)
    return output_list, ratio_list


@contour_filter
def triangle_filter(contour_list, min_area_ratio=0.8, approximation_coefficient=0.02):
    """
    Receives a list of contours and returns only those that are approximately
    triangle and have 3
    :param approximation_coefficient: the approximation coefficient affects
    how the contour's perimeter is approximated, the larger the number is (out of 1) the more the approximation of
    the contour is a line
    :param contour_list: the list of contours to be filtered
    :param min_area_ratio: the minimum ratio between the area of the contour and the bounding triangle
    :return: the filtered list
    """
    output_list = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio = triangle_fill_ratio(current_contour)
        peri = cv2.arcLength(current_contour, True)
        approximation = cv2.approxPolyDP(current_contour, approximation_coefficient * peri, True)
        if fill_ratio > min_area_ratio and len(approximation) == 3:
            ratio_list.append(fill_ratio)
            output_list.append(current_contour)
    return output_list, ratio_list


@contour_filter
def circle_filter(contour_list, min_area_ratio=0.7):
    """
    Filters out contour which are not approximately circle.
    :param contour_list: list of contours (numpy arrays) to be filtered
    :param min_area_ratio: minimum ratio between the area of the enclosing
                         circle and the contour (contour /enclosing circle)
    :return: the list of contours that fit the criteria
    """
    output = []
    ratio_list = []
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, radius = circle_fill_ratio(current_contour)
        if min_area_ratio <= fill_ratio:
            output.append(current_contour)
            ratio_list.append(fill_ratio)
    return output, ratio_list


@contour_filter
def constraining_circle_filter(contour_list, min_area_ratio=0.80, min_len_ratio=0.9):
    """
    Filters out contour which are not approximately circle, also limits by radius ratio (unlike circle_filter).
    :param contour_list: list of contours (numpy arrays) to be filtered
    :param min_area_ratio: minimum ratio between the area of the enclosing
                      circle and the contour (contour /enclosing circle)
    :param min_len_ratio: minimum ratio between the radius of the enclosing circle and the enclosing rectangle.
    :return: the list of contours that fit the criteria
    """
    output = []
    ratios = []
    output_append = output.append
    ratio_append = ratios.append
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        fill_ratio, radius = circle_fill_ratio(current_contour)
        _, _, rectangle_width, rectangle_height = cv2.boundingRect(current_contour)
        radius_ratio = ((2 * radius) ** 2) / float(rectangle_width) * rectangle_height
        if min_len_ratio <= (radius_ratio ** 0.5):
            if min_area_ratio <= fill_ratio:
                output_append(current_contour)
                ratio_append((fill_ratio, radius_ratio))
    return output, ratios


@contour_filter
def polygon_filter(contour_list, side_amount=6, min_angle_ratio=0.7,
                   min_area_ratio=0.7, min_len_ratio=0.7, approximation_coefficient=0.02):
    """
     A filter that Detects regular polygon of n-sides sides
    :param contour_list: the list of contours to be filtered
    :param side_amount: the amount of sides the wanted polygon has.
    :param min_angle_ratio: the minimum ratio between the each angle and the average angle and the average angle and the
                            target angle of a shape of side_amount
    :param min_area_ratio: The minimum ratio between the contour's area and the target area
    :param min_len_ratio: The minimum ratio between the length of each side and the average length and
                          between the average length.
    :param approximation_coefficient: the coefficient for the function cv2.approxPolyDP.
    :return: the list of contours (numpy ndarrays) that passed the filter and are regular polygon
    """
    if side_amount < 3:
        raise ValueError("A polygon must have at least 3 sides")
    output = []
    ratios = []
    output_append = output.append
    ratio_append = ratios.append
    if type(contour_list) is not list:
        contour_list = [contour_list]
    for current_contour in contour_list:
        vertices, lengths, angles = contour_lengths_and_angles(current_contour, approximation_coefficient)
        if len(vertices) == side_amount:
            target_angle = regular_polygon_angle(side_amount)
            average_length = round(sum(lengths) / float(len(lengths)), 3)
            average_angle = round(sum(angles) / float(len(lengths)), 3)
            for length in lengths:
                if 1 - ((((length - average_length) ** 2) ** 0.5) / average_length) < min_len_ratio:
                    break
            else:
                for angle in angles:
                    if 1 - ((((angle - average_angle) ** 2) ** 0.5) / average_angle) < min_angle_ratio:
                        break
                else:
                    if not 1 - ((((target_angle - average_angle) ** 2) ** 0.5) / target_angle) < min_angle_ratio:
                        ratios = cv2.contourArea(current_contour) / polygon_area(average_length, side_amount)
                        if min_area_ratio <= ratios <= 1 / min_area_ratio:
                            ratio_append(ratios)
                            output_append(current_contour)
    return output, ratios


@contour_filter
def percent_area_filter(contour_list, minimal_percent=0.002, maximum_percent=1, image_dimensions=(320, 240)):
    """
    Filters out contours that are not in the specified ratio to the area of the image (1% -> 0.1)
    :param contour_list: list of contours to be filtered (numpy.ndarray)
    :param minimal_percent: the minimal ratio between the contour area and the image area
    :param maximum_percent: the maximum ratio between the contour area and the image area
    :param image_dimensions: The (width, height) of the image in pixels,
    :return:
    """
    output = []
    ratios = []
    output_append = output.append
    ratio_append = ratios.append
    image_size = image_dimensions[0] * image_dimensions[1]
    if image_size == 0:
        raise ValueError("Invalid image dimensions, Received (width, height): {}, {}".format(*image_dimensions))
    for current_contour in contour_list:
        contour_area = cv2.contourArea(current_contour)
        percent_area = contour_area / image_size
        if minimal_percent <= percent_area <= maximum_percent:
            ratio_append(percent_area)
            output_append(current_contour)
    return output, ratios



def function_description(function):
    return ("Function Name: {function_name}\n"
            "documentation:\n{documentation}"
            .format(function_name=function.__name__, documentation=function.__doc__))



import cv2

from .naming_conventions import *


def photo_array(camera, amount, delay=3, name_convention=time_name_convention, path=None):
    """
     Takes a series of images with a delay between shots and saves them,
            after every image, press 'y' to save it and 'n' to pass
    :param camera: the camera to take images with, ovl.Camera or cv2.VideoCapture
    :param amount: the amount of images to take, must be a positive integer
    :param delay: the delay between images
    :param name_convention: the naming convention of the images
    :param path: if images should be saved in a specific folder
    :return: the images
    """
    images = []
    for image_number in range(amount):
        ret, image = camera.read()
        if ret:
            images.append(image)
            save_name = name_convention(image_number)
            save_name = path + save_name if path else save_name
            while True:
                cv2.imshow('Ovl Photo Array', image)
                key = cv2.waitKey(delay)
                if key == ord('y'):
                    cv2.imwrite(save_name, image)
                    break
                elif key == ord('n'):
                    break
    return images



# Copyright 2018-2019 Ori Ben-Moshe - All rights reserved.
from numpy import ndarray

from ..math_ import geometry


def xy_center_directions(contours, image: ndarray):
    """
    Returns the director for the robot for the given contours for the x and y axis
    :param contours: the final contours found after filtering - your targets
    :param image: the image from which it was found
    :return: the director
    """
    height, width, _ = image.shape
    x, y = geometry.contour_average_center(contours)
    x /= width / 2
    y /= height / 2
    return x - 1, y - 1


def y_center_directions(contour, image: ndarray):
    """
    Returns the director for the robot for the given contours for the y axis only
    the value of the direction is between -1 and 1
    :param contour: the final contours - your targets
    :param image: the image from which it was found
    :return: the direction to move in the y axis (-1 to 1)
    """
    height, _, _ = image.shape
    _, y = geometry.contour_average_center(contour)
    y /= height / 2
    return y - 1


def x_center_directions(contour, image: ndarray):
    """
    Returns the director for the robot for the given contours for the x axis only
    :param contour: the final contours - your targets
    :param image: the image from which it was found
    :return: the direction to move in the y axis (-1 to 1)
    """
    _, width, _ = image.shape
    x, _ = geometry.contour_average_center(contour)
    x /= width / 2
    return x - 1


def center_directions(contour, image: ndarray):
    """
    Returns the average center of the contours
    or list of contours that are the final targets

    This is the default directions function since it
    doesnt calculate any directions, only finds the center
    :param contour: the final contours - your targets
    :param image: the image from which it was found
    """
    return geometry.contour_average_center(contour)



from typing import Union


class DirectionMonitor:
    @property
    def priority(self) -> Union[bool, float]:
        return False

    def monitor(self, directions, contours, image, mask):
        raise NotImplementedError("When inheriting DirectionMonitor you must implement it!")



import numpy as np


class CameraSettings:
    def __init__(self, camera_offset, camera_calibration):
        self.camera_offset = camera_offset
        self.camera_calibration = camera_calibration
        self.image_dimensions = None

    def undistort_image(self, image: np.ndarray) -> np.ndarray:
        """
         Removes distortion created by imperfections in the camera.
        :param image: The image (numpy ndarray) that should be undistorted.
        :return: an undistorted copy of the image
        """
        return self.camera_calibration.undistort_image(image)



from setuptools import setup, find_packages

with open("README.md", "r") as read_me:
    long_description = read_me.read()

setup(
    name='ovl',
    packages=find_packages(),
    version='2020.1.4',
    license='apache-2.0',
    author='Ori Ben-Moshe',
    author_email='ovl.contact.help@gmail.com',
    description='A modular and versatile Python package for Computer Vision'
                ' object detection pipelines Tailored for robotics applications',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/1937Elysium/Ovl-Python',
    install_requires=['numpy', 'pynetworktables', 'opencv-python'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows'
    ]
)



from typing import Any, Union

from .connection import Connection
from ..helpers_.team_number_to_ip import team_number_to_ip

NetworkTables = None


class NetworkTablesConnection(Connection):
    """
    Note: In Order to use NetworkTablesConnection you must have pynetworktables installed.

    A connection to NetworkTables (The FRC network protocol)
    NetworkTables are a group of Dictionaries (Hash tables) that are shared by all computers in the network.
    NetworkTablesConnection creates defaults to writing to the Vision table.
    You can then read from /vision/vision_result the result sent.
    For Additional information about NetworkTables
    please refer to:
    https://docs.wpilib.org/en/latest/docs/software/networktables/networktables-intro.html
    """

    def __init__(self, roborio: str, table_name: str = "Vision", table_key: str = "vision_result"):
        """
        :param roborio: team number, roborio hostname or ip.
            for a team number 1234 valid inputs are:
              - 1234
              - roboRIO-1234-FRC.local
              - 10.12.34.2 (10.TE.AM.2)
              - a dynamic ip of the roborio
        NOTE: When using static ip (relevant inputs: 1234 and 10.12.34.2)
              make sure both your computer and the roborio are set to have a static ip
              for more information:
              https://docs.wpilib.org/en/latest/docs/networking/networking-introduction/index.html
        :param table_name: the table to use as a default "Vision" is recommended
        :param table_key: the key in the table to use as a default
        """
        from networktables import NetworkTables
        if roborio.isnumeric():
            roborio = team_number_to_ip(roborio)
        self.roborio_ip = roborio
        self.connection = NetworkTables.initialize(roborio)
        self.table_key = table_key
        self.table = NetworkTables.getTable(table_name)
        self.table_cache = {table_name: self.table}
        self.table_name = table_name

    def get_table(self, table):
        """
        Fetches a table by name, if table exists in cache it returns
        the cached table instead.
        If table is none it returns the default table (set in the constructor)
        :param table: the name of the table
        :return: the table
        """
        if table is None:
            table = self.table
        elif table in self.table:
            table = self.table_cache[table]
        else:
            table_name = table
            table = NetworkTables.getTable(table)
            self.table_cache[table_name] = table
        return table

    def send(self, data, table_key: Union[str, None] = None,
             table: Union[str, None] = None, *args, **kwargs) -> None:
        """
        A function used to put a value in a given table and key in the connected
        NetworkTable.
        For more information about NetworkTables functionality
        please refer to:
        https://docs.wpilib.org/en/latest/docs/software/networktables/networktables-intro.html
        :param data: the data to send (post)
        :param table_key: the specific table to read from
        :param table: the table to receive from. Examples are SmartDashboard
                      or Usage.Use "Vision" if you are not sure)
        """
        table_key = table_key if table_key else self.table_key
        table = self.get_table(table)
        return table.putValue(table_key, data)

    def receive(self, table_key: str = None, table: str = None,
                default_value=None, *args, **kwargs) -> Any:
        """
        Gets a value from network tables from which to read a value
        :param table_key: the specific table to read from
        :param table: the table to receive from (Use "Vision" if you are not sure)
        :param default_value: the value to return if the given table key does not exist
        """
        table = self.get_table(table)
        table_key = table_key or self.table_key
        return table.getValue(table_key, default_value)



import numpy as np
from typing import List, Tuple
import types

from . import vision
from ..connections import connection
from ..thresholds.threshold import Threshold


class AmbientVision:
    """
    A Mash of 2 vision objects that extends one of the visions at a time.
    The visions are swapped every main_amount of update_vision calls.
    This makes the "ambient_vision" run in the background once every multiple frames
    """

    def __init__(self, main_vision: vision.Vision, ambient_vision: vision.Vision,
                 main_amount: int, start_ambient: bool = False):
        self._main_vision = main_vision
        self._ambient_vision = ambient_vision
        self.main_amount = main_amount
        self._counter = 0
        self._current_vision = main_vision
        self._is_ambient = start_ambient

    @property
    def main_vision(self) -> vision.Vision:
        return self.main_vision

    @property
    def ambient_vision(self) -> vision.Vision:
        return self.ambient_vision

    @property
    def connection(self) -> connection.Connection:
        return self._current_vision.connection

    @property
    def threshold(self) -> Threshold:
        return self._current_vision.threshold

    @property
    def image_filters(self) -> List[types.FunctionType]:
        return self._current_vision.image_filters

    @property
    def directions(self):
        return self._current_vision.director

    def is_ambient(self) -> bool:
        """
        Returns true if the current vision is the ambient one
        """
        return self._is_ambient

    def get_image(self) -> np.ndarray:
        """
         Take a picture using the current vision
        """
        return self._current_vision.get_image()

    def get_filtered_image(self) -> np.ndarray:
        """
         Returns the image from the current vision with
                all of the image filters applied on it.
        :return: the filtered image
        :rtype: ndarray
        """
        return self._current_vision.get_filtered_image()

    def apply_image_filters(self, image: np.ndarray) -> np.ndarray:
        """
        Applies all of the image filter of the current vision on the given image
        """
        return self._current_vision.apply_image_filters(image)

    def apply_morphological_functions(self, mask: np.ndarray) -> np.ndarray:
        """
        Applies the current vision's morphological functions on
        a binary mask (numpy ndarray)
        See Vision.apply_morphological_functions for full description
        """
        return self._current_vision.apply_morphological_functions(mask)

    def find_contours_in_mask(self, mask: np.ndarray, save: bool = True,
                              return_hierarchy: bool = False) -> List[np.ndarray]:
        """
        Returns a list of contours from a given binary mask (numpy ndarray),
        using the current vision
        See Vision.find_contours_in_mask for full description
        """
        return self._current_vision.find_contours_in_mask(mask, save, return_hierarchy)

    def find_contours(self, image: np.ndarray, threshold=None, return_hierarchy=False) -> List[np.ndarray]:
        """
        Find contours in the given image, using the current vision.
        See Vision.find_contours for full description
        """
        return self._current_vision.find_contours(image, threshold, return_hierarchy)

    def detect(self, image: np.ndarray, verbose=False
               ) -> Tuple[List[np.ndarray], List[float], np.ndarray]:
        """
        Gets contours and applies all filters and returns the result,
        thus detecting the object according to the specifications in the vision,
        Uses the current vision
        :param image: image in which the vision should detect an object
        :param verbose: if true prints additional information about contour filtering
        :return: contours, ratio list (from the filter functions) and the filtered image
        """
        return self._current_vision.detect(image, verbose=verbose)

    def update_vision(self):
        """
        Increases the inner counter and swaps the ambient and the main vision
        after the set number of updates (self.main_amount)
        """
        if self._counter < self.main_amount:
            self._counter += 1
            self._current_vision = self.main_vision
            self._is_ambient = False
        else:
            self._counter = 0
            self._current_vision = self.ambient_vision
            self._is_ambient = True
